@@ -13,19 +13,28 @@ use MIME::Base64 qw(encode_base64url);
 use WWW::Mechanize;
 use URI;
 
-our $VERSION = '0.01';
+our $VERSION = '0.02';
 
 my $home_dir;
 
+# The %api_cache hash is a cache for Tesla API call data across all objects.
+# The $api_cache_time is a timestamp of last cache write for a particular
+# endpoint/ID pair, and is relative to API_CACHE_TIMEOUT_SECONDS
+
+my %api_cache;
+my $api_cache_time = time;
+
 BEGIN {
     $home_dir = File::HomeDir->my_home;
 }
 
 use constant {
-    CACHE_FILE  => "$home_dir/tesla_api_cache.json",
-    AUTH_URL    => 'https://auth.tesla.com/oauth2/v3/authorize',
-    TOKEN_URL   => 'https://auth.tesla.com/oauth2/v3/token',
-    API_URL     => 'https://owner-api.teslamotors.com/',
+    DEBUG_CACHE                 => 0,
+    API_CACHE_TIMEOUT_SECONDS   => 2,
+    CACHE_FILE                  => "$home_dir/tesla_api_cache.json",
+    AUTH_URL                    => 'https://auth.tesla.com/oauth2/v3/authorize',
+    TOKEN_URL                   => 'https://auth.tesla.com/oauth2/v3/token',
+    API_URL                     => 'https://owner-api.teslamotors.com/',
 };
 
 # Public object methods
@@ -43,14 +52,15 @@ sub new {
         return $self;
     }
 
+    $self->api_cache_time($params{api_cache_time});
+
     $self->mech;
     $self->_access_token;
 
-    $self->my_vehicle_id($params{vehicle_id});
-
     return $self;
 }
 sub api {
+
     my ($self, $endpoint_name, $id) = @_;
 
     if (! defined $endpoint_name) {
@@ -70,6 +80,18 @@ sub api {
         $uri =~ s/\{.*?\}/$id/;
     }
 
+    # Return early if all cache mechanisms check out
+
+    if ($self->api_cache_time) {
+        if (time - $api_cache_time <= $self->api_cache_time) {
+            if ($self->_cache(endpoint => $endpoint_name, id => $id)) {
+                print "Returning cache for $endpoint_name/$id pair...\n" if DEBUG_CACHE;
+                return $self->_cache(endpoint => $endpoint_name, id => $id);
+            }
+        }
+        print "No cache present for $endpoint_name/$id pair...\n" if DEBUG_CACHE;
+    }
+
     my $url = URI->new(API_URL . $uri);
 
     my $request;
@@ -86,15 +108,28 @@ sub api {
     my $response = $self->mech->request($request);
 
     if ($response->is_success) {
-        return _decode($response->decoded_content)->{response};
+        my $response_data = _decode($response->decoded_content)->{response};
+
+        $self->_cache(
+            endpoint => $endpoint_name,
+            id       => $id,
+            data     => $response_data
+        );
+
+        return $response_data;
     }
     else {
         warn $response->status_line;
     }
 }
-sub data {
-    my ($self) = @_;
-    return $self->{data};
+sub api_cache_time {
+    my ($self, $cache_seconds) = @_;
+
+    if (defined $cache_seconds) {
+        $self->{api_cache_time} = $cache_seconds;
+    }
+
+    return $self->{cache_api_data} // API_CACHE_TIMEOUT_SECONDS;
 }
 sub endpoints {
     my ($self, $endpoint) = @_;
@@ -134,54 +169,9 @@ sub mech {
 
     $self->{mech} = $www_mech;
 }
-sub my_vehicle_id {
-    my ($self, $id) = @_;
-
-    if (defined $id) {
-        $self->{data}{vehicle_id} = $id;
-    }
-    else {
-        my @vehicle_ids = keys %{$self->my_vehicles};
-
-        if (scalar @vehicle_ids == 1) {
-            $self->{data}{vehicle_id} = $vehicle_ids[0];
-        }
-    }
-
-    return $self->{data}{vehicle_id} || -1;
-}
-sub my_vehicle_name {
+sub object_data {
     my ($self) = @_;
-
-    if (! $self->my_vehicle_id) {
-        warn "You haven't set a vehicle ID yet";
-    }
-
-    return $self->my_vehicles->{$self->my_vehicle_id};
-}
-sub my_vehicles {
-    my ($self) = @_;
-
-    return $self->{vehicles} if $self->{vehicles};
-
-    my $vehicles = $self->api('VEHICLE_LIST');
-
-    for (@$vehicles) {
-        $self->{data}{vehicles}{$_->{id}} = $_->{display_name};
-    }
-
-    return $self->{data}{vehicles};
-}
-
-# Public Tesla API methods
-
-sub vehicle_data {
-    my ($self, $id) = @_;
-    return $self->api('VEHICLE_DATA', $self->_id($id));
-}
-sub wake {
-    my ($self, $id) = @_;
-    return $self->api('VEHICLE_DATA', $self->_id($id));
+    return $self->{data};
 }
 
 # Private methods
@@ -192,31 +182,37 @@ sub _access_token {
 
     my ($self) = @_;
 
-    return $self->{access_token} if $self->{access_token};
-
     if (! -e CACHE_FILE) {
         my $auth_code = $self->_authentication_code;
         $self->_access_token_generate($auth_code);
     }
-    else {
-        $self->{access_token} = $self->_access_token_fetch;
+
+    my $valid_token = $self->_access_token_validate;
+
+    if (! $valid_token) {
+        $self->_access_token_refresh;
     }
+
+    $self->{access_token} = $self->_access_token_data->{access_token};
+
+    return $self->{access_token};
 }
-sub _access_token_fetch {
-    # Fetches the access token from the cache file
+sub _access_token_data {
+    # Fetches and stores the cache data file dat
 
-    my ($self) = @_;
+    my ($self, $data) = @_;
+
+    $self->{cache_data} = $data if defined $data;
+
+    return $self->{cache_data} if $self->{cache_data};
 
-    my $cache_data;
     {
-        open my $fh, '<', CACHE_FILE or croak "Can't open Tesla cache file " . CACHE_FILE . ": $!";
+        open my $fh, '<', CACHE_FILE or die "Can't open Tesla cache file " . CACHE_FILE . ": $!";
         my $json = <$fh>;
-        $cache_data = decode_json($json);
+        $self->{cache_data} = decode_json($json);
     }
 
-    my $access_token = $cache_data->{access_token};
-
-    return $access_token;
+    return $self->{cache_data};
 }
 sub _access_token_generate {
     # Generates an access token and stores it in the cache file
@@ -243,18 +239,99 @@ sub _access_token_generate {
     my $response = $self->mech->request($request);
 
     if ($response->is_success) {
-        open my $fh, '>', CACHE_FILE or die $!;
-        print $response->decoded_content;
-        print $fh $response->decoded_content;
+        my $token_data = decode_json($response->decoded_content);
 
-        my $response_data = decode_json($response->decoded_content);
+        $token_data = $self->_access_token_set_expiry($token_data);
+        $self->_access_token_update($token_data);
 
-        return $response_data;
+        return $token_data;
     }
     else {
         croak $self->mech->response->status_line;
     }
 }
+sub _access_token_validate {
+    # Checks the validity of an existing token
+
+    my ($self) = @_;
+
+    my $token_expires_at = $self->_access_token_data->{expires_at};
+    my $token_expires_in = $self->_access_token_data->{expires_in};
+
+    my $valid = 0;
+
+    if (time + $token_expires_in < $token_expires_at) {
+        $valid = 1;
+    }
+
+    return $valid;
+}
+sub _access_token_set_expiry {
+    # Sets the access token expiry date/time after generation and
+    # renewal
+
+    my ($self, $token_data) = @_;
+
+    if (! defined $token_data || ref($token_data) ne 'HASH') {
+        croak "_access_token_set_expiry() needs a hash reference of token data";
+    }
+
+    my $expiry = time + $token_data->{expires_in};
+
+    $token_data->{expires_at} = $expiry;
+
+    return $token_data;
+}
+sub _access_token_refresh {
+    # Renews an expired/invalid access token
+
+    my ($self) = @_;
+
+    my $url = URI->new(TOKEN_URL);
+    my $header = ['Content-Type' => 'application/json; charset=UTF-8'];
+
+    my $refresh_token = $self->_access_token_data->{refresh_token};
+
+    my $request_data = {
+        grant_type    => 'refresh_token',
+        refresh_token => $refresh_token,
+        client_id     => 'ownerapi',
+    };
+
+    my $request = HTTP::Request->new('POST', $url, $header, encode_json($request_data));
+
+    my $response = $self->mech->request($request);
+
+    if ($response->is_success) {
+        my $token_data = decode_json($response->decoded_content);
+
+        # Re-add the existing refresh token; its still valid
+        $token_data->{refresh_token} = $refresh_token;
+
+        # Set the expiry time
+        $token_data = $self->_access_token_set_expiry($token_data);
+
+        # Update the cached token
+        $self->_access_token_update($token_data);
+    }
+    else {
+        croak $self->mech->response->status_line;
+    }
+}
+sub _access_token_update {
+    # Writes the new or updated token to the cache file
+
+    my ($self, $token_data) = @_;
+
+    if (! defined $token_data || ref($token_data) ne 'HASH') {
+        croak "_access_token_update() needs a hash reference of token data";
+    }
+
+    $self->_access_token_data($token_data);
+
+    open my $fh, '>', CACHE_FILE or die $!;
+    print $fh encode_json($token_data);
+}
 sub _authentication_code {
     # If an access token is unavailable, prompt the user with a URL to
     # authenticate to Tesla, and have them paste in the resulting URL
@@ -278,11 +355,9 @@ sub _authentication_code {
     $auth_url->query_form(%params);
 
     print
-        qq~
-            Please follow the URL displayed below in your browser and log into Tesla,
-            then paste the URL from the resulting "Page Not Found" page's address bar,
-            then hit ENTER:\n";
-        ~;
+        "Please follow the URL displayed below in your browser and log into Tesla, " .
+        "then paste the URL from the resulting 'Page Not Found' page's address bar, " .
+        "then hit ENTER:\n";
 
     print "\n$auth_url\n";
 
@@ -300,7 +375,6 @@ sub _authentication_code {
         croak "Could not extract the authorization code from the URL";
     }
 
-    print "$code\n";
     return $code;
 }
 sub _authentication_code_verifier {
@@ -319,6 +393,25 @@ sub _authentication_code_verifier {
 
     return $self->{authentication_code_verifier} = $code_verifier;
 }
+sub _cache {
+    # Stores the Tesla API fetched data
+    my ($self, %params) = @_;
+
+    my $endpoint = $params{endpoint};
+    my $id = $params{id} // 0;
+    my $data = $params{data};
+
+    if (! $endpoint) {
+        croak "_cache() requires an endpoint name sent in";
+    }
+
+    if ($data) {
+        $api_cache{$endpoint}{$id} = $data;
+        $api_cache_time = time;
+    }
+
+    return $api_cache{$endpoint}{$id};
+}
 sub _decode {
     # Decode JSON to Perl
     my ($json) = @_;
@@ -340,21 +433,6 @@ sub _useragent_string {
     my $ua = 'Mozilla/5.0 (Macintosh; Intel Mac OS X 10.15; rv:98.0) Gecko/20100101 Firefox/98.0';
     return $ua;
 }
-sub _id {
-    # Tries to figure out the ID to use in API calls
-
-    my ($self, $id) = @_;
-
-    if (! defined $id) {
-        $id = $self->my_vehicle_id;
-    }
-
-    if (! $id) {
-        croak "vehicle_data() requires an \$id sent in";
-    }
-
-    return $id;
-}
 
 1;
 
@@ -378,31 +456,24 @@ Tesla::API - Interface to Tesla's API
 
     my @endpoint_names = keys %{ $tesla->endpoints };
 
-    # Using the internal api() until the complete interface
-    # of this distribution is done
-
     my $endpoint_name = 'VEHICLE_DATA';
 
-    my $car_data = $tesla->api($endpoint_name, $tesla->vehicle_id);
+    my $car_data = $tesla->api($endpoint_name, $tesla->id);
 
 =head1 DESCRIPTION
 
 This distribution provides access to the Tesla API.
 
-B<WARNING>: This is an initial, beta release. Barely any functionality has
-been implemented, and the authentication mechanism needs a lot of polishing.
+B<WARNING>: This is an initial, beta release. The interface *WILL* change.
 
-It's currently in its infancy, so the interface may^H^H^Hwill change. Although
-there are very few public access methods available yet, all current and future
-ones behave the exact same way, by using the object's C<api()> method with an
-endpoint name.
+This class is designed to be subclassed. For example, I have already begun a
+new L<Tesla::Vehicle> distribution which will have access and update methods
+that deal specifically with Tesla autos, then a C<Tesla::Powerwall>
+distribution for their battery storage etc.
 
-Some endpoints require an ID sent in, so it must be provided for those calls as
-well.
-
-B<< NOTE >>: The 'wake' function has not yet been fully impemented, so if a
-Tesla API call times out, its likely you'll have to use the official Tesla
-App to wake the car up.
+Some endpoints require an ID sent in, so it must be provided for those calls
+as well. Some endpoints also require additional parameters for setting
+attributes, but I haven't got that far yet.
 
 =head1 METHODS
 
@@ -424,36 +495,34 @@ B<NOTE>: If you do not have a Tesla account, you can still instantiate a
 L<Tesla::API> object by supplying the C<< unauthenticated => 1 >> parameter
 to C<new()>.
 
-Parameters:
+B<Parameters>:
 
 All parameters are to be sent in the form of a hash.
 
     unauthenticated
 
-Optional, Bool: Set to true to bypass the access token generation.
+I<Optional, Bool>: Set to true to bypass the access token generation.
 
-Default: False.
+I<Default>: C<undef>
 
-    vehicle_id
+    api_cache_time
 
-Optional, Integer: If sent in, we'll use this ID for all calls to endpoints
-that require one.
+I<Optional, Integer>: By default, we cache the fetched data from the Tesla API
+for two seconds. If you make calls that have already been called within that
+time, we will return the cached data.
 
-If not sent in, we'll check how many vehicles you own under your account, and
-if there's only one, we'll use that ID instead.
+Send in the number of seconds you'd like to cache the data for. A value of zero
+(C<0>) will disable caching and all calls through this library will go directly
+to Tesla every time.
 
-If you have more than one Tesla vehicle registered and you don't supply this
-parameter, you will have to supply the ID to each method call that requires it,
-or set it in C<vehicle_id($id)> after instantiation.
-
-Default: C<undef>
+I<Return>: Integer, the number of seconds we're caching Tesla API data for.
 
 =head2 api($endpoint, $id)
 
 Responsible for disseminating the endpoints and retrieving data through the
 Tesla API.
 
-Parameters:
+B<Parameters>:
 
     $endpoint
 
@@ -462,15 +531,35 @@ found in the C<t/test_data/endpoints.json> file for the time being.
 
     $id
 
-Optional, Integer: Some endpoints require an ID sent in (eg. vehicle ID,
-powerwall ID etc).
+I<Optional, Integer>: Some endpoints require an ID sent in (eg. vehicle ID,
+Powerwall ID etc).
 
-Return: Hash or array reference, depending on the endpoint.
+I<Return>: Hash or array reference, depending on the endpoint.
 
-=head2 data
+=head2 api_cache_time($cache_seconds)
+
+The number of seconds we will cache retrieved endpoint data from the Tesla API
+for, to reduce the number of successive calls to retrieve the same data.
+
+B<Parameters>:
+
+    $cache_seconds
+
+I<Optional, Integer>: By default, we cache the fetched data from the Tesla API
+for two seconds. If you make calls that have already been called within that
+time, we will return the cached data.
+
+Send in the number of seconds you'd like to cache the data for. A value of zero
+(C<0>) will disable caching and all calls through this library will go directly
+to Tesla every time.
+
+I<Return>: Integer, the number of seconds we're caching Tesla API data for.
+
+=head2 object_data
 
 Returns a hash reference of the data we've collected for you and stashed
-within the object. This does not reflect the entire object.
+within the object. This does not reflect the entire object, just the data
+returned from Tesla's API.
 
 =head2 endpoints
 
@@ -502,70 +591,23 @@ To get a list of endpoint names:
 
 =head2 mech
 
-Returns the L<WWW::Mechanze> object we've instantiated internally.
-
-=head2 my_vehicle_id($id)
-
-Sets/gets your primary vehicle ID. If set, we will use this in all API calls
-that require it.
-
-Parameters:
-
-    $id
-
-Optional, Integer: The vehicle ID you want to use in all API calls that require
-one, as opposed to sending it into every separate call.
-
-If you only have a single Tesla vehicle registered under your account, we will
-set C<my_vehicle_id()> to that ID when you instantiate the object.
-
-You can also have this auto-populated in C<new()> by sending it in with the
-C<< vehicle_id => $id >> parameter.
-
-=head2 my_vehicle_name
-
-Returns the name you associated with your vehicle under your Tesla account.
-
-L</my_vehicle_id($id)> must have already been set.
-
-=head2 my_vehicles
-
-Returns a hash reference of your listed vehicles. The key is the vehicle ID,
-and the value is the name you've assigned to that vehicle.
-
-Example:
-
-    {
-        1234567891011 => 'Dream Machine',
-        1234567891012 => 'Model S',
-    }
-
-=head2 vehicle_data($id)
-
-Returns a hash reference containing state data of a vehicle.
-
-C<croak()>s if an ID isn't sent in and we can't sort one out automatically.
-
-=head2 wake($id)
-
-NOT YET IMPLEMENTED FULLY.
+Returns the L<WWW::Mechanize> object we've instantiated internally.
 
 =head2 EXAMPLE USAGE
 
+See L<Tesla::Vehicle> for vehicle specific methods.
+
     use Data::Dumper;
     use Tesla::API;
     use feature 'say';
 
     my $tesla = Tesla::API->new;
+    my $vehicle_id = 1234238782349137;
 
-    say $tesla->my_vehicle_name;
-
-    print Dumper $tesla->vehicle_data;
+    print Dumper $tesla->api('VEHICLE_DATA', $vehicle_id);
 
 Output (massively and significantly snipped for brevity):
 
-    Dream machine
-
     $VAR1 = {
         'vehicle_config' => {
             'car_type' => 'modelx',
@@ -660,6 +702,23 @@ Output (massively and significantly snipped for brevity):
 
 Steve Bertrand, C<< <steveb at cpan.org> >>
 
+=head1 ACKNOWLEDGEMENTS
+
+This distribution suite has been a long time in the works. For my other projects
+written in Perl previous to writing this code that required data from the Tesla
+API, I wrapped L<Tim Dorssers|https://github.com/tdorssers> wonderful
+L<TeslaPy|https://github.com/tdorssers/TeslaPy> Python project.
+
+Much of the code in this distribution is heavily influenced by the code his
+project, and currently, we're using a direct copy of its
+L<Tesla API endpoint file|https://github.com/tdorssers/TeslaPy/blob/master/teslapy/endpoints.json>.
+
+Thanks Tim, and great work!
+
+Also thanks goes out to L<https://teslaapi.io>, as a lot of the actual request
+parameter information and response data layout I learned from that site while
+implementing the actual REST calls to the Tesla API.
+
 =head1 LICENSE AND COPYRIGHT
 
 Copyright 2022 Steve Bertrand.
@@ -670,6 +729,10 @@ copy of the full license at:
 
 L<http://www.perlfoundation.org/artistic_license_2_0>
 
+The copied endpoint code data borrowed from Tim's B<TeslaPy> project has been
+rebranded with the Perl license here, as permitted by the MIT license TeslaPy
+is licensed under.
+
 =cut
 
 __DATA__
